/**
 * airconsole.js - an RFC2217 Websocket client for Airconsole
 * Copyright (c) 2014 Cloudstore Ltd
 * http://www.get-console.com/airconsole
 *
 */

function Airconsole() {
    this.websocket = null;
    this._baudrate = 9600;
    this._databits = 8;
    this._parity = "N";
    this._stopbits = 1;
    this._flow = "";
    this.defaultIP = "192.168.10.1";
    this.defaultPort = 8080;
    this.defaultSerialPortNumber = 1;
    this._serialPortNumber = 1;
    
    this.options = [
                    {option: Airconsole.options.BINARY, send: Airconsole.commands.WILL, nsend: Airconsole.commands.WONT, ack: Airconsole.commands.DO, nack: Airconsole.commands.DONT, initial: Airconsole.optstates.REQUESTED, state: Airconsole.optstates.INACTIVE},
                    {option: Airconsole.options.BINARY, send: Airconsole.commands.DO, nsend: Airconsole.commands.DONT, ack: Airconsole.commands.WILL, nack: Airconsole.commands.WONT, initial: Airconsole.optstates.REQUESTED, state: Airconsole.optstates.INACTIVE}
    ];
}

Object.defineProperty(Airconsole.prototype, "baudrate",
                      {
                      get: function() { return this._baudrate; }
                      });
Object.defineProperty(Airconsole.prototype, "databits",
                      {
                      get: function() { return this._databits; }
                      });
Object.defineProperty(Airconsole.prototype, "parity",
                      {
                      get: function() { return this._parity; }
                      });
Object.defineProperty(Airconsole.prototype, "stopbits",
                      {
                      get: function() { return this._stopbits; }
                      });
Object.defineProperty(Airconsole.prototype, "flow",
                      {
                      get: function() { return this._flow; }
                      });
Object.defineProperty(Airconsole.prototype, "dtr",
                      {
                      get: function() { return this._dtr; }
                      });
Object.defineProperty(Airconsole.prototype, "rts",
                      {
                      get: function() { return this._rts; }
                      });
Object.defineProperty(Airconsole.prototype, "msr",
                      {
                      get: function() { return this._msr; }
                      });
Object.defineProperty(Airconsole.prototype, "prevmsr",
                      {
                      get: function() { return this._prevmsr; }
                      });
Object.defineProperty(Airconsole.prototype, "serialPortNumber",
					  {
					  get: function() { return this._serialPortNumber; }
					  });

Airconsole.prototype.connect = function(ip, port, portNumber) {
    // IP and port are both optional - use defaults
    this.ip = (typeof ip === "undefined") ? this.defaultIP : ip;
    this.port = (typeof port === "undefined") ? this.defaultPort : port;
    this._serialPortNumber = (typeof portNumber === "undefined") ? this.defaultSerialPortNumber : portNumber;
    
    var uri = "ws://" + this.ip + ":" + this.port + "/";
    var protocol = "telnetcpcd-" + this._serialPortNumber;
    
    this.websocket = new WebSocket(uri, protocol);
    this.websocket.binaryType = "arraybuffer";
    this.websocket.onopen = this.wsOpen.bind(this);
    this.websocket.onclose = this.wsClose.bind(this);
    this.websocket.onmessage = this.wsMessage.bind(this);
    this.websocket.onerror = this.wsError.bind(this);
}

Airconsole.prototype.disconnect = function() {
    if (this.websocket != null) {
        this.websocket.close();
        this.websocket = null;
    }
}

Airconsole.prototype.telnetSend = function(array) {
    if (this.websocket != null) {
        if (array.length > 0) {
            var binMode = this.webin;
            
            // Do a quick check to see if we can dump straight to network
            var okToWrite = true;
            for (var x=0; x < array.length; x++) {
                var c = array[x];
                if (c == Airconsole.chars.IAC || (c == Airconsole.chars.CR && !binMode)) {
                    okToWrite = false;
                    break;
                }
            }
            
            if (okToWrite) {
                this.websocket.send(new Uint8Array(array));
            } else {
                var buffer = [];
                var offset = 0;
                for (var x=0; x < array.length; x++) {
                    var c = array[x];
                    if (c == Airconsole.chars.IAC) {
                        buffer.push(Airconsole.chars.IAC);
                        buffer.push(Airconsole.chars.IAC);
                    } else if (c == Airconsole.chars.CR && !binMode) {
                        buffer.push(Airconsole.chars.CR);
                        buffer.push(Airconsole.chars.NUL);
                    } else {
                        buffer.push(c);
                    }
                }
                this.websocket.send(new Uint8Array(buffer));
            }
        }
    } else {
        this.notifyError("Not connected");
    }
}

Airconsole.prototype.sendOption = function(cmd, option) {
    if (this.websocket != null) {
        var b = [];
        b.push(Airconsole.chars.IAC);
        b.push(cmd);
        b.push(option);
        this.websocket.send(new Uint8Array(b));
    } else {
        this.notifyError("Not connected");
    }
}

Airconsole.prototype.telnetCharSubOption = function(option, suboption, value) {
    if (this.websocket != null) {
        var b = [];
        b.push(Airconsole.chars.IAC);
        b.push(Airconsole.commands.SB);
        b.push(option);
        b.push(suboption);
        if (value == 0xFF) {
            b.push(Airconsole.chars.IAC);
        }
        b.push(value);
        b.push(Airconsole.chars.IAC);
        b.push(Airconsole.commands.SE);
        this.websocket.send(new Uint8Array(b));
    } else {
        this.notifyError("Not connected");
    }
}

Airconsole.prototype.telnetIntSubOption = function(option, suboption, value) {
    if (this.websocket != null) {
        var b = [];
        var a0 = (value & 0xFF000000) >> 24;
        var a1 = (value & 0x00FF0000) >> 16;
        var a2 = (value & 0x0000FF00) >> 8;
        var a3 = (value & 0x000000FF);
        
        b.push(Airconsole.chars.IAC);
        b.push(Airconsole.commands.SB);
        b.push(option);
        b.push(suboption);
        if (a0 == 0xFF) {
            b.push(Airconsole.chars.IAC);
        }
        b.push(a0);
        if (a1 == 0xFF) {
            b.push(Airconsole.chars.IAC);
        }
        b.push(a1);
        if (a2 == 0xFF) {
            b.push(Airconsole.chars.IAC);
        }
        b.push(a2);
        if (a3 == 0xFF) {
            b.push(Airconsole.chars.IAC);
        }
        b.push(a3);
        b.push(Airconsole.chars.IAC);
        b.push(Airconsole.commands.SE);
        this.websocket.send(new Uint8Array(b));
    } else {
        this.notifyError("Not connected");
    }
}

Airconsole.prototype.sendBinary = function(data) {
    if (this.websocket != null) {
        this.telnetSend(data);
    } else {
        this.notifyError("Not connected");
    }
}

Airconsole.prototype.decodeUTF8 = function(bytes) {
    var len = bytes.length;
    var result = "";
    var code;
    var i;
    for (i = 0; i < len; i++) {
        if (bytes[i] <= 0x7f) {
            result += String.fromCharCode(bytes[i]);
        } else if (bytes[i] >= 0xc0) {                                   // Mutlibytes
            if (bytes[i] < 0xe0) {                                       // 2 bytes
                code = ((bytes[i++] & 0x1f) << 6) | (bytes[i] & 0x3f);
            } else if (bytes[i] < 0xf0) {                                // 3 bytes
                code = ((bytes[i++] & 0x0f) << 12) | ((bytes[i++] & 0x3f) << 6)  | (bytes[i] & 0x3f);
            } else {                                                     // 4 bytes
                // turned into two characters in JS as surrogate pair
                code = (((bytes[i++] & 0x07) << 18) |
                        ((bytes[i++] & 0x3f) << 12) |
                        ((bytes[i++] & 0x3f) << 6) |
                        (bytes[i] & 0x3f)) - 0x10000;
                // High surrogate
                result += String.fromCharCode(((code & 0xffc00) >>> 10) + 0xd800);
                // Low surrogate
                code = (code & 0x3ff) + 0xdc00;
            }
            result += String.fromCharCode(code);
        } else {
            // Invalid UTF-8. Most likely this is a cp1252 or iso-8859-1 character in a UTF-8 stream...!
            // Skipped
        }
    }
    return result;
}

Airconsole.prototype.encodeUTF8 = function(str) {
    var utf8 = [];
    for (var i=0; i < str.length; i++) {
        var charcode = str.charCodeAt(i);
        if (charcode < 0x80) {
            utf8.push(charcode);
        } else if (charcode < 0x800) {
            utf8.push(0xc0 | (charcode >> 6),
                      0x80 | (charcode & 0x3f));
        } else if (charcode < 0xd800 || charcode >= 0xe000) {
            utf8.push(0xe0 | (charcode >> 12),
                      0x80 | ((charcode>>6) & 0x3f),
                      0x80 | (charcode & 0x3f));
        } else {
            // surrogate pair
            i++;
            // UTF-16 encodes 0x10000-0x10FFFF by
            // subtracting 0x10000 and splitting the
            // 20 bits of 0x0-0xFFFFF into two halves
            charcode = 0x10000 + (((charcode & 0x3ff)<<10)
                                  | (str.charCodeAt(i) & 0x3ff))
            utf8.push(0xf0 | (charcode >>18),
                      0x80 | ((charcode>>12) & 0x3f),
                      0x80 | ((charcode>>6) & 0x3f),
                      0x80 | (charcode & 0x3f));
        }
    }
    return utf8;
}

Airconsole.prototype.sendText = function(str) {
    // TODO: Allow selection of character encoding - this is UTF-8 only
    var uintArray = this.encodeUTF8(str);
    this.sendBinary(uintArray);
}

Airconsole.prototype.sendHexBytes = function(str) {
    var bytes = [];
    str = str.trim();
    while (str.length >= 2) {
        bytes.push(parseInt(str.substring(0, 2), 16));
        str = str.substring(2, str.length).trim();
    }
    this.sendBinary(bytes);
}

Airconsole.prototype.telnetBreak = function(on) {
    if (on) {
        return Airconsole.rfc2217.BREAK_ON;
    } else {
        return Airconsole.rfc2217.BREAK_OFF;
    }
}
Airconsole.prototype.telnetDTR = function(on) {
    if (on) {
        return Airconsole.rfc2217.DTR_ON;
    } else {
        return Airconsole.rfc2217.DTR_OFF;
    }
}
Airconsole.prototype.telnetRTS = function(on) {
    if (on) {
        return Airconsole.rfc2217.RTS_ON;
    } else {
        return Airconsole.rfc2217.RTS_OFF;
    }
}


Airconsole.prototype.telnetDataBits = function(configDataBits) {
    switch (configDataBits) {
        case 8:
        default:
            return Airconsole.rfc2217.DATASIZE_8;
        case 7:
            return Airconsole.rfc2217.DATASIZE_7;
    }
}
Airconsole.prototype.configDataBits = function(telnetDataBits) {
    switch (telnetDataBits) {
        case Airconsole.rfc2217.DATASIZE_8:
        default:
            return 8;
        case Airconsole.rfc2217.DATASIZE_7:
            return 7;
    }
}

Airconsole.prototype.telnetParity = function(configParity) {
    if (configParity == "N") {
        return Airconsole.rfc2217.PARITY_NONE;
    } else if (configParity == "O") {
        return Airconsole.rfc2217.PARITY_ODD;
    } else if (configParity == "E") {
        return Airconsole.rfc2217.PARITY_EVEN;
    } else if (configParity == "M") {
        return Airconsole.rfc2217.PARITY_MARK;
    } else if (configParity == "S") {
        return Airconsole.rfc2217.PARITY_SPACE;
    } else {
        return Airconsole.rfc2217.PARITY_NONE;
    }
}

Airconsole.prototype.configParity = function(telnetParity) {
    switch (telnetParity) {
        case Airconsole.rfc2217.PARITY_NONE:
        default:
            return "N";
        case Airconsole.rfc2217.PARITY_ODD:
            return "O";
        case Airconsole.rfc2217.PARITY_EVEN:
            return "E";
        case Airconsole.rfc2217.PARITY_MARK:
            return "M";
        case Airconsole.rfc2217.PARITY_SPACE:
            return "S";
    }
}

Airconsole.prototype.telnetStopBits = function(configStopBits) {
    switch (configStopBits) {
        case Airconsole.rfc2217.STOPBITS_1:
        default:
            return 1;
        case Airconsole.rfc2217.STOPBITS_2:
            return 2;
    }
}

Airconsole.prototype.configStopBits = function(telnetStopBits) {
    switch (telnetStopBits) {
        case 1:
        default:
            return Airconsole.rfc2217.STOPBITS_1;
        case 2:
            return Airconsole.rfc2217.STOPBITS_2;
    }
}

Airconsole.prototype.telnetFlowControl = function(configFlowControl) {
    if (configFlowControl == "HW" || configFlowControl == "RTS" || configFlowControl == "CTS") {
        return Airconsole.rfc2217.FLOW_HARDWARE;
    } else if (configFlowControl == "SW" || configFlowControl == "XON" || configFlowControl == "XOFF") {
        return Airconsole.rfc2217.FLOW_SOFTWARE;
    } else {
        return Airconsole.rfc2217.FLOW_NONE;
    }
}

Airconsole.prototype.configFlowControl = function(telnetFlowControl) {
    switch (telnetFlowControl) {
        case Airconsole.rfc2217.FLOW_HARDWARE:
            return "HW";
        case Airconsole.rfc2217.FLOW_SOFTWARE:
            return "SW";
        default:
            return "";
    }
}

Airconsole.prototype.sendBreak = function() {
    if (this.websocket != null) {
        this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_CONTROL, this.telnetBreak(true));
        this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_CONTROL, this.telnetBreak(false));
    } else {
        this.notifyError("Not connected");
    }
}

Airconsole.prototype.setDTR = function(flag) {
    if ((flag == "false") || (flag == "0")) {
        flag = false;
    }
    this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_CONTROL, this.telnetDTR(flag));
}

Airconsole.prototype.setRTS = function(flag) {
    if ((flag == "false") || (flag == "0")) {
        flag = false;
    }
    this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_CONTROL, this.telnetRTS(flag));
}

Airconsole.prototype.telnetInit = function () {
    this.webin = false;
    this.theybin = false;
    this.state = Airconsole.states.TOP_LEVEL;
    this.subopt = 0;
    this.suboptbuf = new Array();

    var index;
    for (index=0; index < this.options.length; index++) {
        var option = this.options[index];
        option.state = option.initial;
        if (option.state == Airconsole.optstates.REQUESTED) {
            this.sendOption(option.send, option.option);
        }
    }
    
    this._msr = 0;
    this._prevmsr = 0;
    
    this._dtr = false;
    this._rts = false;
    
    
    // Set defaults
    this.setLineParameters(this._baudrate, this._databits, this._parity, this._stopbits);
    this.setFlowControl(this._flow);
    
    var initialDTR = false;
    var initialRTS = false;
    if (this.telnetFlowControl(this._flow) == Airconsole.rfc2217.FLOW_HARDWARE) {
        initialDTR = true;
        initialRTS = true;
    }
    
    this.setDTR(initialDTR);
    this.setRTS(initialRTS);
    this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_CONTROL, this.telnetBreak(false));
    this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.PURGE_DATA, Airconsole.rfc2217.PURGE_RXTX);
    
    // Enable notifications for all modem status changes
    this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_MODEMSTATE_MASK, 0xFF);
    this.notifyModemPropertiesChanged();
}

Airconsole.prototype.setLineParameters = function (baudrate, databits, parity, stopbits) {
    if (this.websocket != null) {
        this.telnetIntSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_BAUDRATE, baudrate);
        this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_DATASIZE, this.telnetDataBits(databits));
        this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_PARITY, this.telnetParity(parity));
        this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_STOPSIZE, this.telnetStopBits(stopbits));
    } else {
        // Parameter validation via round-tripping
        this._baudrate = parseInt(baudrate, 10);
        this._databits = this.configDataBits(this.telnetDataBits(databits));
        this._parity = this.configParity(this.telnetParity(parity));
        this._stopbits = this.configStopBits(this.telnetStopBits(stopbits));
    }
}

Airconsole.prototype.setFlowControl = function (flow) {
    if (this.websocket != null) {
        this.telnetCharSubOption(Airconsole.options.COM_PORT_OPTION, Airconsole.rfc2217.SET_CONTROL, this.telnetFlowControl(flow));
    } else {
        // User has passed "config" value - round trip it
        this._flow = this.configFlowControl(this.telnetFlowControl(flow));
    }
}

Airconsole.prototype.wsOpen = function (evt) {
    this.telnetInit();
    if (typeof this.onConnect === 'function') {
        this.onConnect();
    }
}

Airconsole.prototype.wsClose = function (evt) {
    if (typeof this.onDisconnect === 'function') {
        this.onDisconnect();
    }
}

Airconsole.prototype.notifyError = function (message) {
    if (typeof this.onError === 'function') {
        this.onError(message);
    }
}

Airconsole.prototype.notifyLinePropertiesChanged = function () {
    if (typeof this.onLinePropertiesChanged === 'function') {
        this.onLinePropertiesChanged();
    }
}


Airconsole.prototype.notifyModemPropertiesChanged = function () {
    if (typeof this.onModemPropertiesChanged === 'function') {
        this.onModemPropertiesChanged();
    }
}

Airconsole.prototype.hexArray = function (array) {
    var hexEncodeArray = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'];
    var s = '';
    for (var i = 0; i < array.length; i++) {
        var code = array[i];
        s += "0x";
        s += hexEncodeArray[code >>> 4];
        s += hexEncodeArray[code & 0x0F];
        s += " ";
    }
    return s;
}

Airconsole.prototype.readInt = function(array, offset) {
    return (array[offset] << 24) | (array[offset+1] << 16) | (array[offset+2] << 8) | (array[offset+3]);
}

Airconsole.prototype.activateOption = function(option) {
    this.optionSideEffects(option, true);
}

Airconsole.prototype.deactivateOption = function(option) {
    if ((option.state == Airconsole.optstates.REQUESTED) || (option.state == Airconsole.optstates.ACTIVE)) {
        this.sendOption(option.nsend, option.option);
    }
    option.state = Airconsole.optstates.REALLY_INACTIVE;

}

Airconsole.prototype.refusedOption = function(option) {
    this.optionSideEffects(option, false);
}

Airconsole.prototype.optionSideEffects = function(option, enabled) {
    if (option.option == Airconsole.options.BINARY && option.send == Airconsole.commands.DO) {
        this.theybin = enabled;
    } else if (option.option == Airconsole.options.BINARY && option.send == Airconsole.commands.WILL) {
        this.webin = enabled;
    }
}

Airconsole.prototype.processRxOption = function(cmd, option) {
    var index;
    var o;
    for (index = 0; index < this.options.length; index++) {
        o = this.options[index];
        if ((o.option == option) && (o.ack == cmd)) {
            switch (o.state) {
                case Airconsole.optstates.REQUESTED:
                    o.state = Airconsole.optstates.ACTIVE;
                    this.activateOption(o);
                    break;
                case Airconsole.optstates.ACTIVE:
                    break;
                case Airconsole.optstates.INACTIVE:
                    o.state = Airconsole.optstates.ACTIVE;
                    this.sendOption(o.send, option);
                    this.activateOption(o);
                    break;
                case Airconsole.optstates.REALLY_INACTIVE:
                    this.sendOption(o.nsend, option);
                    break;
            }
            return;
        } else if ((o.option == option) && (o.nack == cmd)) {
            switch (o.state) {
                case Airconsole.optstates.REQUESTED:
                    o.state = Airconsole.optstates.INACTIVE;
                    this.refusedOption(o);
                    break;
                case Airconsole.optstates.ACTIVE:
                    o.state = Airconsole.optstates.INACTIVE;
                    this.sendOption(o.nsend, option);
                    this.optionSideEffects(o, false);
                    break;
                case Airconsole.optstates.INACTIVE:
                case Airconsole.optstates.REALLY_INACTIVE:
                    break;
            }
            return;
        }
    }
    if (cmd == Airconsole.commands.WILL || cmd == Airconsole.commands.DO) {
        this.sendOption((cmd == Airconsole.commands.WILL) ? Airconsole.commands.DONT : Airconsole.commands.WONT, option);
    }
}

Airconsole.prototype.processSubOption = function() {
    switch (this.subopt) {
        case Airconsole.options.COM_PORT_OPTION:
            if (this.suboptbuf.length >= 1) {
                var comopt = this.suboptbuf[0];
                var optlen = this.suboptbuf.length-1;
                var configValue;

                switch (comopt) {
                    case Airconsole.rfc2217.NOTIFY_MODEMSTATE + Airconsole.rfc2217.SERVER_PREFIX:
                        if (optlen == 1) {
                            var value = this.suboptbuf[1];
                            if (this._msr != value) {
                                this._prevmsr = this._msr;
                                this._msr = value;
                                this.notifyModemPropertiesChanged();
                            }
                        }
                        break;
                        
                    case Airconsole.rfc2217.SET_MODEMSTATE_MASK + Airconsole.rfc2217.SERVER_PREFIX:
                        if (optlen == 1) {
                            var value = this.suboptbuf[1];
                            // modemstate mask set to value
                        }
                        break;
                        
                    case Airconsole.rfc2217.SET_CONTROL + Airconsole.rfc2217.SERVER_PREFIX:
                        if (optlen == 1) {
                            var value = this.suboptbuf[1];
                            switch (value) {
                                case Airconsole.rfc2217.DTR_ON:
                                    if (!this._dtr) {
                                        this._dtr = true;
                                        this.notifyLinePropertiesChanged();
                                    }
                                    break;
                                case Airconsole.rfc2217.DTR_OFF:
                                    if (this._dtr) {
                                        this._dtr = false;
                                        this.notifyLinePropertiesChanged();
                                    }
                                    break;
                                case Airconsole.rfc2217.RTS_ON:
                                    if (!this._rts) {
                                        this._rts = true;
                                        this.notifyLinePropertiesChanged();
                                    }
                                    break;
                                case Airconsole.rfc2217.RTS_OFF:
                                    if (this._rts) {
                                        this._rts = false;
                                        this.notifyLinePropertiesChanged();
                                    }
                                    break;
                                case Airconsole.rfc2217.BREAK_ON:
                                    break;
                                case Airconsole.rfc2217.BREAK_OFF:
                                    break;
                                case Airconsole.rfc2217.FLOW_NONE:
                                case Airconsole.rfc2217.FLOW_SOFTWARE:
                                case Airconsole.rfc2217.FLOW_HARDWARE:
                                case Airconsole.rfc2217.FLOW_HARDWARE_DSR:
                                    configValue = this.configFlowControl(value);
                                    if (this._flow != configValue) {
                                        this._flow = configValue;
                                        this.notifyLinePropertiesChanged();
                                    }
                                    break;
                                default:
                                    break;
                            }
                            
                        }
                        break;
                        
                        
                    case Airconsole.rfc2217.SET_BAUDRATE + Airconsole.rfc2217.SERVER_PREFIX:
                        if (optlen == 4) {
                            var value = this.readInt(this.suboptbuf, 1);
                            if (value > 0) {
                                if (this._baudrate != value) {
                                    this._baudrate = value;
                                    this.notifyLinePropertiesChanged();
                                }
                            }
                        }
                        break;
                    case Airconsole.rfc2217.SET_DATASIZE + Airconsole.rfc2217.SERVER_PREFIX:
                        if (optlen == 1) {
                            var value = this.suboptbuf[1];
                            configValue = this.configDataBits(value);
                            if (this._databits != configValue) {
                                this._databits = configValue;
                                this.notifyLinePropertiesChanged();
                            }
                        }
                        break;
                    case Airconsole.rfc2217.SET_PARITY + Airconsole.rfc2217.SERVER_PREFIX:
                        if (optlen == 1) {
                            var value = this.suboptbuf[1];
                            configValue = this.configParity(value);
                            if (this._parity != configValue) {
                                this._parity = configValue;
                                this.notifyLinePropertiesChanged();
                            }
                        }
                        break;
                    case Airconsole.rfc2217.SET_STOPSIZE + Airconsole.rfc2217.SERVER_PREFIX:
                        if (optlen == 1) {
                            var value = this.suboptbuf[1];
                            configValue = this.configStopBits(value);
                            if (this._stopbits != configValue) {
                                this._stopbits = configValue;
                                this.notifyLinePropertiesChanged();
                            }
                        }
                        break;
                    default:
                        break;
                }
                
            }
            break;

    }
}

Airconsole.prototype.processTelnetData = function (array) {

    var buffer = [];
    
    for (var x=0; x < array.length; x++) {
        var c = array[x];
        switch (this.state) {
            case Airconsole.states.TOP_LEVEL:
            case Airconsole.states.SEENCR:
                if (c == Airconsole.chars.NUL && this.state == Airconsole.states.SEENCR) {
                    this.state = Airconsole.states.TOP_LEVEL;
                } else if (c == Airconsole.chars.IAC) {
                    this.state = Airconsole.states.SEENIAC;
                } else {
                    buffer.push(c);
                    if (c == Airconsole.chars.CR && !this.theybin) {
                        this.state = Airconsole.states.SEENCR;
                    } else {
                        this.state = Airconsole.states.TOP_LEVEL;
                    }
                }
                break;
                
            case Airconsole.states.SEENIAC:
                if (c == Airconsole.commands.DO) {
                    this.state = Airconsole.states.SEENDO;
                } else if (c == Airconsole.commands.DONT) {
                    this.state = Airconsole.states.SEENDONT;
                } else if (c == Airconsole.commands.WILL) {
                    this.state = Airconsole.states.SEENWILL;
                } else if (c == Airconsole.commands.WONT) {
                    this.state = Airconsole.states.SEENWONT;
                } else if (c == Airconsole.commands.SB) {
                    this.state = Airconsole.states.SEENSB;
                } else {
                    if (c == Airconsole.chars.IAC) {
                        buffer.push(c);
                    }
                    this.state = Airconsole.states.TOP_LEVEL;
                }
                break;
                
            case Airconsole.states.SEENWILL:
                this.processRxOption(Airconsole.commands.WILL, c);
                this.state = Airconsole.states.TOP_LEVEL;
                break;
            case Airconsole.states.SEENWONT:
                this.processRxOption(Airconsole.commands.WONT, c);
                this.state = Airconsole.states.TOP_LEVEL;
                break;
            case Airconsole.states.SEENDO:
                this.processRxOption(Airconsole.commands.DO, c);
                this.state = Airconsole.states.TOP_LEVEL;
                break;
            case Airconsole.states.SEENDONT:
                this.processRxOption(Airconsole.commands.DONT, c);
                this.state = Airconsole.states.TOP_LEVEL;
                break;
            case Airconsole.states.SEENSB:
                this.subopt = c;
                this.suboptbuf = new Array();
                this.state = Airconsole.states.SUBNEGOT;
                break;
            case Airconsole.states.SUBNEGOT:
                if (c == Airconsole.chars.IAC) {
                    this.state = Airconsole.states.SUBNEG_IAC;
                } else {
                    this.suboptbuf.push(c);
                }
                break;
            case Airconsole.states.SUBNEG_IAC:
                if (c != Airconsole.commands.SE) {
                    this.suboptbuf.push(c);
                    this.state = Airconsole.states.SUBNEGOT;
                } else {
                    this.processSubOption();
                    this.state = Airconsole.states.TOP_LEVEL;
                }
                break;
        }
    }
    
    if (buffer.length > 0) {
        if (typeof this.onData === 'function') {
            this.onData(new Uint8Array(buffer));
        }
    }
}

Airconsole.prototype.dataRead = function(reader) {
    var array = new Uint8Array(reader.result);
    this.processTelnetData(array);
}

Airconsole.prototype.wsMessage = function (evt) {
    if (evt.data instanceof ArrayBuffer) {
        var array = new Uint8Array(evt.data);
        this.processTelnetData(array);
    } else if (evt.data instanceof Blob) {
        var b = evt.data;
        var reader = new FileReader();
        reader.addEventListener("loadend", this.dataRead.bind(this, reader));
        reader.readAsArrayBuffer(b);
    } else {
        this.notifyError("Received unknown message type. Event="+evt+", data="+evt.data);
    }
}

Airconsole.prototype.wsError = function (evt) {
    this.notifyError("Websocket error: " + evt);
}

Airconsole.states = Object.freeze({
    TOP_LEVEL:0,
    SEENIAC:1,
    SEENWILL:2,
    SEENWONT:3,
    SEENDO:4,
    SEENDONT:5,
    SEENSB:6,
    SUBNEGOT:7,
    SUBNEG_IAC:8,
    SEENCR:9
});

Airconsole.commands = Object.freeze({
    DONT:254,
    DO:253,
    WONT:252,
    WILL:251,
    SB:250,
    SE:240
});

Airconsole.optstates = Object.freeze({
    REQUESTED:0,
    ACTIVE:1,
    INACTIVE:2,
    REALLY_INACTIVE:3
});

Airconsole.chars = Object.freeze({
    CR: 13,
    LF: 10,
    NUL: 0,
    IAC: 255
});

Airconsole.options = Object.freeze({
    BINARY:0,
    ECHO: 1,
    SGA:3,
    NAWS: 31,
    TSPEED: 32,
    COM_PORT_OPTION: 44
});

Airconsole.rfc2217 = Object.freeze({
    COM_PORT_SIGNATURE:0,
    SET_BAUDRATE:1,
    SET_DATASIZE: 2,
    SET_PARITY: 3,
    SET_STOPSIZE: 4,
    SET_CONTROL: 5,
    NOTIFY_LINESTATE: 6,
    NOTIFY_MODEMSTATE: 7,
    SET_LINESTATE_MASK: 10,
    SET_MODEMSTATE_MASK: 11,
    PURGE_DATA: 12,
    SERVER_PREFIX: 100,
                                   
    DATASIZE_5: 5,
    DATASIZE_6: 6,
    DATASIZE_7: 7,
    DATASIZE_8: 8,
                                   
    PARITY_NONE: 1,
    PARITY_ODD: 2,
    PARITY_EVEN: 3,
    PARITY_MARK: 4,
    PARITY_SPACE: 5,
                                   
    STOPBITS_1: 1,
    STOPBITS_2: 2,
    STOPBITS_15: 3,
                                   
    BREAK_QUERY: 4,
    BREAK_ON: 5,
    BREAK_OFF: 6,
                                   
    DTR_QUERY: 7,
    DTR_ON: 8,
    DTR_OFF: 9,
                                   
    RTS_QUERY: 10,
    RTS_ON: 11,
    RTS_OFF: 12,
                                   
    FLOW_QUERY: 0,
    FLOW_NONE: 1,
    FLOW_SOFTWARE: 2,
    FLOW_HARDWARE: 3,
    FLOW_HARDWARE_DCD: 17,
    FLOW_HARDWARE_DSR: 19,
                                   
    MODEM_DCD: 0x80,
    MODEM_RI: 0x40,
    MODEM_DSR: 0x20,
    MODEM_CTS: 0x10,
    MODEM_DDCD: 0x08,
    MODEM_DRI: 0x04,
    MODEM_DDSR: 0x02,
    MODEM_DCTS: 0x01,
                                   
    PURGE_RX: 1,
    PURGE_TX: 2,
    PURGE_RXTX: 3
});
